/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2017/18      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : tutorial 2
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: race conditions
///
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <vector>
#include <thread>
#include <cstdlib>
#include <cmath>

std::atomic<int> sum;
std::mutex summutex;

////////////////////////////////////////////////////////////////////////////////////////////////////
// add two vectors together
////////////////////////////////////////////////////////////////////////////////////////////////////
void
compute_norm( int tid, int num_threads, int* a, int n, int& sum) {

  int chunk_size = n / num_threads;
  int i1 = tid * chunk_size;
  int i2 = (tid+1) * chunk_size;

  for( int i = i1; i < i2; ++i) {
    sum += a[i] * a[i];
  }

#if 0
  summutex.lock();
  sum += suml;
  summutex.unlock();
#endif
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  const unsigned int n = 8192 * 8192;
  const unsigned int num_threads = std::thread::hardware_concurrency() / 2;

  // allocate memory
  int* a = (int*) malloc( n * sizeof(int));
  for( unsigned int i = 0; i < n; i++) {
    a[i] = 1;
  }

  sum = 0;

  // create threads
  std::vector< std::thread > threads;
  std::vector<int> sums( num_threads, 0);
  for( unsigned int i = 0; i < num_threads; ++i) {
    threads.push_back( std::thread( compute_norm, i, num_threads, a, n,
                                    std::ref( sums[i]) ));
  }

 for( unsigned int i = 0; i < num_threads; ++i) {
    threads[i].join();
    sum += sums[i];
  }

  std::cerr << "norm(a) = " << std::sqrt( (float) sum) << std::endl;

  // cleanup memory
  free( a);

  return EXIT_SUCCESS;
}
